from td import root, COMP, baseCOMP, op  # pyright: ignore[reportMissingImports]
from os import environ
from pathlib import Path

def ensure_global_tox(
		filepath, 
		op_shortcut, 
		root_comp:COMP = root, 
		default_path = "utils", 
		reloadcustom = False, 
		reloadbuiltin = False):

	if (_potentialy:= getattr(op, op_shortcut, None)) is not None:
		return _potentialy

	current_comp = root_comp
	for path_element in environ.get("ENSURE_UTILITY_PATH", default_path ).strip("/ ").split("/"):
		current_comp = current_comp.op( path_element ) or current_comp.create( baseCOMP, path_element)

	newly_loaded 							= current_comp.loadTox(filepath)
	newly_loaded.name 						= op_shortcut
	newly_loaded.par.opshortcut.val 		= op_shortcut
	newly_loaded.par.externaltox.val 		= Path(filepath).relative_to( Path(".").absolute() )
	newly_loaded.par.enableexternaltox.val 	= True
	newly_loaded.par.savebackup.val 		= False
	newly_loaded.par.reloadcustom.val 		= reloadcustom
	newly_loaded.par.reloadbuiltin.val 		= reloadbuiltin
	newly_loaded.par.enableexternaltoxpulse.pulse()
	return newly_loaded
	
from typing import TypeVar, Type, cast
T = TypeVar("T")
def ensure_global_tdp( tdp_module, default_path = "utils", root_comp:COMP = root, cast_as:Type[T] = COMP ):
	return  cast(T,
		ensure_global_tox(
				 tdp_module.ToxFile,
				 getattr( tdp_module, "DefaultGlobalOpShortcut", tdp_module.__name__.capitalize() ),
				 root_comp      = root_comp,
				 default_path   =   default_path
             )
    )





from .partypes import ParMode


import typing as _T
ParValueT 	= _T.TypeVar('ParValueT')
ParTypeT 	= _T.TypeVar("ParTypeT")

from typing import NotRequired, TypedDict, Tuple, Any, Literal
from abc import abstractmethod, abstractproperty
###
class _ParGroupArgs( TypedDict, _T.Generic[ParValueT] ):
		#size        : Literal[2,3,4]
		label 		: NotRequired[str]
		mode 		: NotRequired[Tuple[ParMode,...]]
		defaultMode : NotRequired[Tuple[ParMode,...]]
		default     : NotRequired[Tuple[ParValueT,...]]
		defaultExpr : NotRequired[Tuple[str, ...]]
		defaultBindExpr : NotRequired[Tuple[str,...]]		
		expr        : NotRequired[Tuple[str,...]]
		bindExpr    : NotRequired[Tuple[str,...]]
		
		readOnly    : NotRequired[bool]
		help        : NotRequired[str]
		enable      : NotRequired[bool]
		enableExpr  : NotRequired[str]

class _ParGroup(_T.Generic[ParValueT, ParTypeT]):
	class _args(_ParGroupArgs[ParValueT]): # pyright: ignore[reportGeneralTypeIssues]
		pass

	@abstractmethod
	def __getitem__(self, index:Literal[2,3,4]) -> ParTypeT:
		pass

	defaultMode : Tuple[ParMode,...]
	mode		: Tuple[ParMode,...]
	
	_creationMethodName : str
	style:str

	val:Tuple[ParValueT,...]
	
	@abstractmethod
	def eval(self) -> Tuple[ParValueT,...]:
		pass
	
	default : Tuple[ParValueT,...]
	defaultExpr : Tuple[str,...]
	defaultBindExpr : Tuple[str,...]
	
	readOnly : bool
	enable : bool

	owner : Any

	expr : Tuple[str,...]
	enableExpr : Tuple[str,...]
	bindExpr :Tuple[str,...]
	
	name : str
	label : str
	help : str
	
	@abstractmethod
	def destroy(self):
		pass
	@abstractmethod
	def reset() -> bool:
		pass
	@abstractmethod
	def isPar( par:Any ) -> bool:
		pass

class _NumericParGroupArgs( TypedDict, _T.Generic[ParValueT] ):	
		min : NotRequired[Tuple[ParValueT,...]]
		max : NotRequired[Tuple[ParValueT,...]]
		normMin : NotRequired[Tuple[ParValueT,...]]
		normMax : NotRequired[Tuple[ParValueT,...]]
		clampMin : NotRequired[Tuple[ParValueT,...]]
		clampMax : NotRequired[Tuple[ParValueT,...]]
		


class _NumericParGroup(_ParGroup[ParValueT, ParTypeT]):
	class _args(_NumericParGroupArgs[ParValueT], _ParGroupArgs[ParValueT]): # pyright: ignore[reportGeneralTypeIssues]
		pass
	min : Tuple[ParValueT,...]
	max : Tuple[ParValueT,...]
	normMin : Tuple[ParValueT,...]
	normMax : Tuple[ParValueT,...]
	clampMin : Tuple[ParValueT,...]
	clampMax : Tuple[ParValueT,...]
	@abstractmethod
	def evalNorm(self) -> Tuple[ParValueT,...]:
		pass


from .partypes import ParFloat, ParInt

class ParGroupFloat(_NumericParGroup["float", ParFloat]):
	class _args(_NumericParGroupArgs[float], _ParGroupArgs[float]):
		size : Literal[2,3,4]
	"TD Float Parameter"
	style:str = "Float"

class ParGroupInt(_NumericParGroup["int", ParInt]):
	class _args(_NumericParGroupArgs[int], _ParGroupArgs[int]):
		size : Literal[2,3,4]
	"TD Int Parameter"
	style:str = "Int"
	
class ParGroupRGBA(_NumericParGroup["float", ParFloat]):
	class _args(_NumericParGroupArgs[float], _ParGroupArgs[float]):
		size : Literal[2,3,4]
	"TD RGBA Parameter"
	style:str = "RGBA"
	
class ParGroupXYZW(_NumericParGroup["float", ParFloat]):
	class _args(_NumericParGroupArgs[float], _ParGroupArgs[float]):
		size : Literal[2,3,4]
	"TD XYZW Parameter"
	style:str = "XYZW"
	
class ParGroupUVW(_NumericParGroup["float", ParFloat]):
	class _args(_NumericParGroupArgs[float], _ParGroupArgs[float]):
		size : Literal[2,3]
	"TD XYZW Parameter"
	style:str = "UVW"
	
class ParGroupWH(_NumericParGroup["float", ParFloat]):
	class _args(_NumericParGroupArgs[float], _ParGroupArgs[float]):
		size : Literal[2]
	"TD XYZW Parameter"
	style:str = "WH"


"""
partypes.py

This file contains special classes for TouchDesigner parameters.
These objects don't exist in TD but are necessary for tdi to 
provide proper return types from parameters.
"""

from typing import Any, List
from abc import abstractmethod
from enum import Enum

class ParMode(Enum):
	BIND = "BIND"
	CONSTANT = "CONSTANT"
	EXPORT = "EXPORT"
	EXPRESSION = "EXPRESSION"


import typing as _T
ParValueT = _T.TypeVar('ParValueT')

from typing import NotRequired, TypedDict
###
class _ParArgs( TypedDict, _T.Generic[ParValueT] ):
		label 		: NotRequired[str]
		mode 		: NotRequired[ParMode]
		defaultMode : NotRequired[ParMode]
		default : NotRequired[ParValueT]
		defaultExpr : NotRequired[str]
		defaultBindExpr : NotRequired[str]
		readOnly : NotRequired[bool]
		enable : NotRequired[bool]
		expr : NotRequired[str]
		enableExpr : NotRequired[str]
		bindExpr : NotRequired[str]
		help : NotRequired[str]

class _Par(_T.Generic[ParValueT]):
	class _args(_ParArgs[ParValueT]): # pyright: ignore[reportGeneralTypeIssues]
		pass

	defaultMode : ParMode
	mode		: ParMode
	
	_creationMethodName : str
	style:str

	val:ParValueT
	@abstractmethod
	def eval(self) -> ParValueT:
		pass
	
	default : ParValueT
	defaultExpr : str
	defaultBindExpr : str
	
	readOnly : bool
	enable : bool

	owner : Any

	expr : str
	enableExpr : str
	bindExpr :str
	
	name : str
	label : str
	help : str
	
	@abstractmethod
	def destroy(self):
		pass
	
	@abstractmethod
	def reset(self) -> bool:
		pass

	@abstractmethod
	def isPar( self, par:Any ) -> bool:
		pass

class _NumericParArgs( TypedDict, _T.Generic[ParValueT] ):	
		min : NotRequired[ParValueT]
		max : NotRequired[ParValueT]
		normMin : NotRequired[ParValueT]
		normMax : NotRequired[ParValueT]
		clampMin : NotRequired[ParValueT]
		clampMax : NotRequired[ParValueT]
		


class _NumericPar(_Par[ParValueT]):
	class _args(_NumericParArgs[ParValueT], _ParArgs[ParValueT]): # pyright: ignore[reportGeneralTypeIssues]
		pass
	min : ParValueT
	max : ParValueT
	normMin : ParValueT
	normMax : ParValueT
	clampMin : ParValueT
	clampMax : ParValueT
	@abstractmethod
	def evalNorm(self) -> ParValueT:
		pass


class _MenuParArgs( TypedDict ):
		menuLabels : NotRequired[List[str]]
		menuNames : NotRequired[List[str]]
		menuSource : NotRequired[str]
		

class _MenuPar( _Par["str"]):
	class _args(_MenuParArgs, _ParArgs["str"]): # pyright: ignore[reportGeneralTypeIssues]
		pass
	menuNames : List[str]
	menuLabels : List[str]
	menuSource : str
	"""
	Get or set an expression that returns an object with .menuItems .menuNames members. This can be used to create a custom menu whose entries dynamically follow that of another menu for example. Simple menu sources include another parameter with a menu c, an object created by tdu.TableMenu, or an object created by TDFunctions.parMenu.
	```
	p.menuSource = "op('audiodevin1').par.device"
	```
	Note the outside quotes, as menuSource is an expression, not an object.
	"""




class ParStr(_Par["str"]):
	"TD Str Parameter"
	style:str = "Str"

class ParFloat(_NumericPar["float"]):
	"TD Float Parameter"
	style:str = "Float"

class ParInt(_NumericPar["int"]):
	"TD Int Parameter"
	style:str = "Int"

class ParToggle(_Par["bool"]):
	"TD Toggle Parameter"
	style:str = "Toggle"

class ParMomentary(_Par["bool"]):
	"TD Momentary Parameter"
	style:str = "Momentary"

class ParPulse(_Par["bool"]):
	"TD Pulse Parameter"
	style:str = "Pulse"

class ParMenu(_MenuPar):
	"TD Menu Parameter"
	style:str = "Menu"

class ParStrMenu(_MenuPar):
	"TD StrMenu Parameter"
	style:str = "StrMenu"

class ParCOMP(_Par["None | COMP"]):
	"TD COMP Parameter"
	style:str = "COMP"

class ParOP(_Par["None | OP"]):
	"TD OP Parameter"
	style:str = "OP"

class ParObject(_Par["None | ObjectCOMP"]):
	"TD Object Parameter"
	style:str = "Object"

class ParSOP(_Par["None | SOP"]):
	"TD SOP Parameter"
	style:str = "SOP"

class ParPOP(_Par["None | POP"]):
	"TD POP Parameter"
	style:str = "POP"

class ParMAT(_Par["None | MAT"]):
	"TD MAT Parameter"
	style:str = "MAT"

class ParCHOP(_Par["None | CHOP"]):
	"TD CHOP Parameter"
	style:str = "CHOP"

class ParTOP(_Par["None | TOP"]):
	"TD TOP Parameter"
	style:str = "TOP" 

class ParDAT(_Par["None | DAT"]):
	"TD DAT Parameter"
	style:str = "DAT"

class ParPanelCOMP(_Par["None | PanelCOMP"]):
	"TD PanelCOMP Parameter"
	style:str = "PanelCOMP"

class ParFile(_Par["str"]):
	"TD File Parameter"
	style = "File"

class ParFolder(_Par["str"]):
	"TD Folder Parameter"
	style = "Folder"

# Not yet implemented.
"""
class ParPython(_Par["Any"]):
	"TD Python Parameter"

# These ones are pargrops and can not be represented in the default value.	


class ParRGB(_NumericPar["float"]):
	"TD RGB Parameter"

class ParRGBA(_NumericPar["float"]):
	"TD RGBA Parameter"

class ParUV(_NumericPar["float"]):
	"TD UV Parameter"

class ParUVW(_NumericPar["float"]):
	"TD UVW Parameter"

class ParWH(_NumericPar["float"]):
	"TD WH Parameter"

class ParXY(_NumericPar["float"]):
	"TD XY Parameter"

class ParXYZ(_NumericPar["float"]):
	"TD XYZ Parameter"

class ParXYZW(_NumericPar["float"]):
	"TD XYZW Parameter"

# These ones are special and honsetyl not worth the headache :)


class ParFileSave(_Par["str"]):
	"TD FileSave Parameter"


class ParHeader(_Par["str"]):
	"TD Header Parameter"

class ParSequence(_NumericPar["int"]):
	"TD Sequence Parameter"

class ParDATAdder(_Par["None"]):
	"TD DATAdder Parameter"
"""


from td import * # pyright: ignore[reportMissingImports]

from functools import lru_cache
from typing import Union

@lru_cache( maxsize = None )
def get_parameter_page(operator:OP, pagename:str, builtin:bool = True, custom:bool = True) -> Union[Page, None]:
	for page in operator.pages * builtin + operator.customPages * custom:
		if page.name == pagename: return page

from typing import List, Dict, Any

class MenuSource:
	def __init__(self, names:List[str], labels:List[str] = []) -> None:
		if labels and len(labels) != len(names):
			raise TypeError("Names and labels need same length.")
		self.names = names
		self.labels = labels or names
	
	@property
	def menuLabels(self):
		return self.labels
	
	@property
	def menuNames(self):
		return self.names
	
def dictkeys_to_source(source_dict:Dict[str, Any], capitalize_labels=False):
    return MenuSource(
		[ key for key in source_dict.keys() ],
		[ key.capitalize() for key in source_dict.keys() ] if capitalize_labels else []
	)
import re

def is_legal_name(name:str):
		"""
		Return a legal par name or None if name has no usable characters.
		Size and style are required to do duplicate checking
		if okName is provided, that name will be allowed. Used for name changes.
		"""
		if name is None:
			return False
		# fix characters
		_name = name.strip().lower().capitalize()  # fix cases
		_name = re.sub('[^A-Za-z0-9]+', '', _name)  # remove special chars
		return _name == name


from typing import Any, TypeVar, Type, overload, ClassVar, cast, List, Union
from abc import abstractmethod
from dataclasses import dataclass



from .par_def import partypes
from .par_def import pargrouptypes



## Utils Start
lookupdict = {}


def _pop_default_kwarsg( target_dict:dict ):
    return {
        key : value for key, value in target_dict.items() if key not in  ["page"]
    }

@dataclass
class _ParProxy():
    data : dict
    par_type : Type[partypes._Par]

    def __call__(self, ownerComp) -> Any:
        target_par = ensure_parameter(
            ownerComp, 
            self.data["name"], 
            self.data["page"], 
            f"append{self.par_type.style}", 
            self.par_type.style
        )  

        for attrname, attrvalue in self.data.items():
            if attrname in ("name", "page"): continue
            setattr( target_par, attrname, attrvalue )

        #One can dram. Pleas derivative senpai.
        #_dependency_object = tdu.Dependency( target_par.eval() )
        #_dependency_object.bindMaster = target_par
        #_dependency_object.callbacks.append( self.data.get("callback", lambda *args: None) )

        return target_par
    
@dataclass
class _ParGroupProxy():
    data : dict
    par_type : Type[partypes._Par]

    def __call__(self, ownerComp) -> Any:
        target_par = ensure_pargroup(
            ownerComp, 
            self.data["name"], 
            self.data["page"], 
            f"append{self.par_type.style}", 
            self.par_type.style
        )  

        for attrname, attrvalue in self.data.items():
            if attrname in ("name", "page"): continue
            setattr( target_par, attrname, attrvalue )
        return target_par


def ensure_page(ownerComp, pagename):
    for page in ownerComp.pages:
        if page.name == pagename: return page
    return ownerComp.appendCustomPage( pagename )

from .parameter import is_legal_name

def ensure_parameter(ownerComp, par_name:str, pagename:str, adder_method_name:str, par_style:str):
    if not is_legal_name( par_name ): raise Exception("Illegal CustomPar Name.")
    page = ensure_page( ownerComp, pagename )
    if (par := ownerComp.par[par_name]) is not None:
        # lets validate the partype itself.
        if par.style != par_style: par.destroy()

    if ownerComp.par[par_name] is None:
        # now lets check if the par already exists, if not    
        getattr(page, adder_method_name)( par_name )
    resulting_par = ownerComp.par[par_name] # This noteably only works with single value parameters!
    resulting_par.page = page
    return resulting_par

def ensure_pargroup(ownerComp, par_name:str, pagename:str, adder_method_name:str, par_style:str):
    if not is_legal_name( par_name ): raise Exception("Illegal CustomPar Name.")
    page = ensure_page( ownerComp, pagename )
    if (parGroup := ownerComp.parGroup[par_name]) is not None:
        # lets validate the partype itself.
        if parGroup.style != par_style: parGroup.destroy()

    if ownerComp.parGroup[par_name] is None:
        # now lets check if the par already exists, if not    
        getattr(page, adder_method_name)( par_name )
    resulting_pargroup = ownerComp.parGroup[par_name]
    resulting_pargroup.page = page
    return resulting_pargroup


## Utils End

from typing import Unpack

T = TypeVar("T") 

### Overlods for typehinting.
@overload
def parfield(field_type:Type[partypes.ParStr], page:str = "Custom", **kwargs:Unpack[ partypes.ParStr._args]) -> partypes.ParStr: 
    pass

@overload
def parfield(field_type:Type[partypes.ParFloat], page:str = "Custom", **kwargs:Unpack[ partypes.ParFloat._args]) -> partypes.ParFloat: 
    pass

@overload
def parfield(field_type:Type[partypes.ParInt], page:str = "Custom",**kwargs:Unpack[ partypes.ParInt._args]) -> partypes.ParInt: 
    pass
@overload
def parfield(field_type:Type[partypes.ParToggle], page:str = "Custom", **kwargs:Unpack[ partypes.ParToggle._args]) -> partypes.ParToggle: 
    pass

@overload
def parfield(field_type:Type[partypes.ParMomentary], page:str = "Custom",**kwargs:Unpack[ partypes.ParMomentary._args ]) -> partypes.ParMomentary: 
    pass

@overload
def parfield(field_type:Type[partypes.ParPulse], page:str = "Custom",**kwargs:Unpack[ partypes.ParPulse._args ]) -> partypes.ParPulse: 
    pass

@overload
def parfield(field_type:Type[partypes.ParMenu], page:str = "Custom",**kwargs:Unpack[ partypes.ParMenu._args]) -> partypes.ParMenu: 
    pass

@overload
def parfield(field_type:Type[partypes.ParStrMenu], page:str = "Custom",**kwargs:Unpack[ partypes.ParStrMenu._args]) -> partypes.ParStrMenu: 
    pass

@overload
def parfield(field_type:Type[partypes.ParOP], page:str = "Custom",**kwargs:Unpack[ partypes.ParOP._args]) -> partypes.ParOP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParCOMP], page:str = "Custom",**kwargs:Unpack[ partypes.ParCOMP._args]) -> partypes.ParCOMP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParObject], page:str = "Custom",**kwargs:Unpack[ partypes.ParObject._args]) -> partypes.ParObject: 
    pass

@overload
def parfield(field_type:Type[partypes.ParSOP], page:str = "Custom",**kwargs:Unpack[ partypes.ParSOP._args]) -> partypes.ParSOP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParPOP], page:str = "Custom",**kwargs:Unpack[ partypes.ParPOP._args]) -> partypes.ParPOP: 
    """2025 or higher. Add vesioncheck"""
    pass

@overload
def parfield(field_type:Type[partypes.ParMAT], page:str = "Custom",**kwargs:Unpack[ partypes.ParMAT._args]) -> partypes.ParMAT: 
    pass

@overload
def parfield(field_type:Type[partypes.ParCHOP], page:str = "Custom",**kwargs:Unpack[ partypes.ParCHOP._args]) -> partypes.ParCHOP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParTOP], page:str = "Custom",**kwargs:Unpack[ partypes.ParTOP._args]) -> partypes.ParTOP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParDAT], page:str = "Custom",**kwargs:Unpack[ partypes.ParDAT._args]) -> partypes.ParDAT: 
    pass

@overload
def parfield(field_type:Type[partypes.ParPanelCOMP], page:str = "Custom",**kwargs:Unpack[ partypes.ParPanelCOMP._args]) -> partypes.ParPanelCOMP: 
    pass

@overload
def parfield(field_type:Type[partypes.ParFile], page:str = "Custom",**kwargs:Unpack[ partypes.ParFile._args]) -> partypes.ParFile: 
    pass

@overload
def parfield(field_type:Type[partypes.ParFolder], page:str = "Custom",**kwargs:Unpack[ partypes.ParFolder._args]) -> partypes.ParFolder: 
    pass

# parGroups
@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupFloat], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupFloat._args]) -> pargrouptypes.ParGroupFloat: 
    pass

@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupInt], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupInt._args]) -> pargrouptypes.ParGroupInt: 
    pass

@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupRGBA], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupRGBA._args]) -> pargrouptypes.ParGroupRGBA: 
    pass

@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupUVW], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupUVW._args]) -> pargrouptypes.ParGroupUVW: 
    pass

@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupWH], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupWH._args]) -> pargrouptypes.ParGroupWH: 
    pass

@overload
def pargroupfield(field_type:Type[pargrouptypes.ParGroupXYZW], page:str = "Custom",**kwargs:Unpack[ pargrouptypes.ParGroupXYZW._args]) -> pargrouptypes.ParGroupXYZW: 
    pass


## Actual Implementation.

from sys import _getframe

def parfield(field_type:Type[T], page:str = "Custom", **kwargs) -> T: 
    pass_args = {
        "page" : page,
        **_pop_default_kwarsg( kwargs )
    }
    return cast( T, _ParProxy(pass_args, field_type) )  # pyright: ignore[reportArgumentType]

from typing import Literal

def pargroupfield( field_type:Type[T], page:str = "Custom", **kwargs) -> T:
    pass_args = {
        "page" : page,
        "size" : kwargs.get("size", 2),
        **_pop_default_kwarsg( kwargs )
    }
    return cast( T, _ParGroupProxy(pass_args, field_type) )  # pyright: ignore[reportArgumentType]


class _empty():
    pass

class EnsureExtension():
    par:ClassVar
    parGroup:ClassVar

    def __init__(self, ownerComp) -> None:
        self.par = getattr( self, "par", _empty)() # pyright: ignore[reportAttributeAccessIssue]
        for attr_name in dir(self.par):
            attr_object = getattr( self.par, attr_name )
            if not isinstance( attr_object, _ParProxy): continue
            # Set name HERE!
            attr_object.data["name"] = attr_name
            setattr( self.par, attr_name, attr_object(ownerComp) )
            

        self.parGroup = getattr( self, "parGroup", _empty)() # pyright: ignore[reportAttributeAccessIssue]

        for attr_name in dir(self.parGroup):
            attr_object = getattr( self.parGroup, attr_name )
            if not isinstance( attr_object, _ParGroupProxy): continue
            # Set name HERE!
            attr_object.data["name"] = attr_name
            setattr( self.parGroup, attr_name, attr_object(ownerComp) )


__all__ = [ "EnsureExtension", "parfield", "partypes" ]



demo = None
if demo:

    class extExample( EnsureExtension ):
        class par:
            Foo = parfield(partypes.ParFloat)
            Bar = parfield(partypes.ParFloat, page ="Different", min = 0, max = 10)
            Baba = parfield( partypes.ParMenu, menuLabels=["Eins", "Zwei", "Drei" ], menuNames=["1", "2", "3"], bindExpr="Hello World" )

        class parGroup:
            Hello = pargroupfield( pargrouptypes.ParGroupRGBA, size = 3 )
            Group = pargroupfield( pargrouptypes.ParGroupRGBA, size = 2)

        def __init__(self, ownerComp) -> None:
            super().__init__(ownerComp)
            self.par.Foo.val = 23

    something = extExample(None)
    something.par.Baba.default = 123 # pyright: ignore[reportAttributeAccessIssue] # Errors
    something.par.Baba.default = "123" # Works!
    something.parGroup.Group.eval()[0]
